import torch
import torch.nn as nn
import torch.nn.functional as F


class DiceLoss(nn.Module):
    def __init__(self):
        super(DiceLoss, self).__init__()

    def forward(self, inputs, targets, smooth=1):
        inputs = torch.sigmoid(inputs)

        inputs = inputs.view(-1)
        targets = targets.view(-1)

        intersection = (inputs * targets).sum()
        dice = (2 * intersection + smooth) / (inputs.sum() + targets.sum() + smooth)

        return 1 - dice


class DiceBCELoss(nn.Module):
    def __init__(self):
        super(DiceBCELoss, self).__init__()

    def forward(self, inputs, targets, smooth=1):
        # inputs = torch.sigmoid(inputs)
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        intersection = (inputs * targets).sum()
        dice_loss = 1 - (2. * intersection + smooth) / (inputs.sum() + targets.sum() + smooth)
        BCE = F.binary_cross_entropy(inputs, targets, reduction='mean')
        Dice_BCE = BCE + dice_loss
        return Dice_BCE



import numpy as np


def convert_to_png_for_saving(img: np.ndarray) -> np.ndarray:
    """
    Convert the array to be suitable for PNG saving with skimage.io.imsave.

    :param img: The image to convert
    :type img: np.ndarray
    :return: The converted image
    :rtype: np.ndarray
    """
    normalized_image = (img - np.min(img)) * (
            255.0 / (np.max(img) - np.min(img)))
    normalized_image = normalized_image.astype('uint8')
    return normalized_image



from .utils import patch


def make_train_val(train_images, val_images, norm, stride, width, output_dir, threshold):
    for train_image in train_images:
        patch(train_image,
              f'{output_dir}/train/',
              norm,
              (width, width),
              stride,
              augment=True,
              threshold=threshold,
              save_image=True,
              remove_background_tiles=False
              )
    for val_image in val_images:
        patch(val_image,
              f'{output_dir}/val/',
              norm,
              (width, width),
              stride,
              augment=False,
              threshold=threshold,
              save_image=True,
              remove_background_tiles=True)



import numpy as np
from skimage.morphology import skeletonize, skeletonize_3d


def cl_score(v, s):
    """[this function computes the skeleton volume overlap]

    Args:
        v ([bool]): [image]
        s ([bool]): [skeleton]

    Returns:
        [float]: [computed skeleton volume intersection]
    """
    return np.sum(v * s) / np.sum(s)


def cl_dice(v_p, v_l):
    """[this function computes the cldice metric]

    Args:
        v_p ([bool]): [predicted image]
        v_l ([bool]): [ground truth image]

    Returns:
        [float]: [cldice metric]
    """
    if len(v_p.shape) == 2:
        tprec = cl_score(v_p, skeletonize(v_l))
        tsens = cl_score(v_l, skeletonize(v_p))
    elif len(v_p.shape) == 3:
        tprec = cl_score(v_p, skeletonize_3d(v_l))
        tsens = cl_score(v_l, skeletonize_3d(v_p))
    return 2 * tprec * tsens / (tprec + tsens)



import torch
import torch.nn as nn
import wandb


class ConvBlock(nn.Module):
    def __init__(self, in_c, out_c):
        super().__init__()

        self.conv1 = nn.Conv2d(in_c, out_c, kernel_size=3, padding=1)
        self.bn1 = nn.BatchNorm2d(out_c)

        self.conv2 = nn.Conv2d(out_c, out_c, kernel_size=3, padding=1)
        self.bn2 = nn.BatchNorm2d(out_c)

        self.relu = nn.ReLU()

    def forward(self, inputs):
        x = self.conv1(inputs)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        return x


class EncoderBlock(nn.Module):
    def __init__(self, in_c, out_c):
        super().__init__()

        self.conv = ConvBlock(in_c, out_c)
        self.pool = nn.MaxPool2d((2, 2))

    def forward(self, inputs):
        x = self.conv(inputs)
        p = self.pool(x)

        return x, p


class DecoderBlock(nn.Module):
    def __init__(self, in_c, out_c):
        super().__init__()

        self.up = nn.ConvTranspose2d(in_c, out_c, kernel_size=2, stride=2, padding=0)
        self.conv = ConvBlock(out_c + out_c, out_c)

    def forward(self, inputs, skip):
        x = self.up(inputs)
        x = torch.cat([x, skip], axis=1)
        x = self.conv(x)

        return x


class VsegModel(nn.Module):
    def __init__(self, pretrained: bool = False, device: torch.device = torch.device('cpu')):
        super().__init__()

        """ Encoder """
        self.e1 = EncoderBlock(1, 64)
        self.e2 = EncoderBlock(64, 128)
        self.e3 = EncoderBlock(128, 256)
        self.e4 = EncoderBlock(256, 512)

        """ Bottleneck """
        self.b = ConvBlock(512, 1024)

        """ Decoder """
        self.d1 = DecoderBlock(1024, 512)
        self.d2 = DecoderBlock(512, 256)
        self.d3 = DecoderBlock(256, 128)
        self.d4 = DecoderBlock(128, 64)

        """ Classifier """
        self.outputs = nn.Conv2d(64, 1, kernel_size=1, padding=0)
        self.output_activation = nn.Sigmoid()

        if pretrained:
            run = wandb.init()
            artifact = run.use_artifact('liom-lab/model-registry/Vessel Segmentation:latest', type='model')
            artifact_dir = artifact.download()
            run.finish()

            state = torch.load(artifact_dir + "/checkpoint.latest.pth", map_location=device)
            self.load_state_dict(state)
            self.to(device)

    def forward(self, inputs):
        s1, p1 = self.e1(inputs)
        s2, p2 = self.e2(p1)
        s3, p3 = self.e3(p2)
        s4, p4 = self.e4(p3)

        b = self.b(p4)

        d1 = self.d1(b, s4)
        d2 = self.d2(d1, s3)
        d3 = self.d3(d2, s2)
        d4 = self.d4(d3, s1)

        outputs = self.outputs(d4)
        outputs = self.output_activation(outputs)
        return outputs



import wandb
from skimage.color import label2rgb
from torch.utils.data import DataLoader

from .data import PatchDataset
from .loss import DiceBCELoss
from .model import VsegModel
from .utils import *


def train(model, loader, optimizer, loss_fn, device):
    # Initialize epoch loss to 0
    epoch_loss = 0.0

    # Put in training mode 
    model.train()

    # Iterate over train-loader
    for x, y in tqdm(loader, desc="Training", leave=False, position=1):
        x = x.to(device)
        y = y.to(device)
        optimizer.zero_grad()
        y_pred = model(x)
        loss = loss_fn(y_pred, y)
        loss.backward()
        optimizer.step()
        epoch_loss += loss.item()

    # Normalize cumulative loss for number of examples
    epoch_loss = epoch_loss / len(loader)
    return epoch_loss, y, y_pred, x


def evaluate(model, loader, loss_fn, device):
    # Initialize epoch loss to 0
    # Added metrics
    epoch_loss = 0.0
    f1 = 0.0
    accuracy = 0.0
    jaccard = 0.0
    recall = 0.0

    # Put in eval mode 
    model.eval()
    with torch.no_grad():
        # Iterate over val-loader
        for x, y in tqdm(loader, desc="Validation", leave=False, position=1):
            x = x.to(device, dtype=torch.float32)
            y = y.to(device, dtype=torch.float32)
            y_pred = model(x)
            loss = loss_fn(y_pred, y)
            epoch_loss += loss.item()
            # metrics
            y_m = y.to("cpu")
            y_m = y_m.numpy()
            y_pred_m = y_pred.to("cpu")
            y_pred_m = y_pred_m.numpy()
            [score_f1, score_recall, score_acc, score_jaccard, score_precision] = calculate_metrics(y_m, y_pred_m)
            f1 += score_f1
            accuracy += score_acc
            jaccard += score_jaccard
            recall += score_recall

        # Normalize cumulative loss for number of examples    
        epoch_loss = epoch_loss / len(loader)
        f1 = f1 / len(loader)
        accuracy = accuracy / len(loader)
        jaccard = jaccard / len(loader)
        recall = recall / len(loader)

    return epoch_loss, y, y_pred, x, f1, accuracy, jaccard, recall


def create_images(x, y, pred, num_images=4):
    y_mask = y.cpu().detach().numpy()
    pred_mask = pred.cpu().detach().numpy()
    x = x.cpu().detach().numpy()
    images = []
    num_images = min(num_images, x.shape[0])
    i = 0
    while len(images) < num_images or i - 1 == x.shape[0]:
        # if np.abs(np.round(x.max(), 5) - np.round(x.min(), 5)) <= 0.001:
        #     continue

        img = mask_image(x, y_mask, pred_mask, i)
        images.append(img)
        i += 1

    return images


def mask_image(x, y_mask, pred_mask, i):
    img = x[i, :, :, :].squeeze()
    img = gray2rgb(img)
    y_mask = y_mask[i, :, :, :].squeeze()
    pred_mask = pred_mask[i, :, :, :].squeeze()

    np.place(y_mask, y_mask >= 0.5, 1)
    np.place(y_mask, y_mask < 0.5, 0)

    diff_mask = (pred_mask - y_mask) * pred_mask
    np.place(diff_mask, diff_mask > 0.5, 1)
    np.place(diff_mask, diff_mask <= 0.5, 0)

    np.place(pred_mask, pred_mask > 0.5, 1)
    np.place(pred_mask, pred_mask <= 0.5, 0)

    diff_mask = diff_mask * 3
    pred_mask = (pred_mask - diff_mask) * 2

    labels = np.max([y_mask, diff_mask, pred_mask], axis=0)
    img = label2rgb(labels, image=img, colors=[[0, 0, 1], [0, 1, 0], [1, 0, 0]], alpha=0.3, bg_label=0,
                    bg_color=None)
    return img


def train_model(dataset_dir="data/patches", dev="cuda", output_train="data/training", learning_rate=0.003673,
                batch_size=35, epochs=62, config=None):
    # Setup training parameters and wandb run
    hyperparameter_defaults = dict(
        batch_size=batch_size,
        learning_rate=learning_rate,
        epochs=epochs)

    # Init wandb
    run = wandb.init(
        project="vseg",
        entity="liom-lab",
        mode="online",
        config=hyperparameter_defaults)

    config = wandb.config

    # Identify the images and labels for training
    train_x = sorted(glob(f'{dataset_dir}/train/images/*'))
    train_y = sorted(glob(f'{dataset_dir}/train/masks/*'))
    val_x = sorted(glob(f'{dataset_dir}/val/images/*'))
    val_y = sorted(glob(f'{dataset_dir}/val/masks/*'))

    data_str = f"Dataset Size:\nTrain: {len(train_x)} - Valid: {len(val_x)}\n"
    print(data_str)

    best_epoch = -1
    create_dir(f"{output_train}")
    create_dir(f"{output_train}/files")
    checkpoint_path = f"{output_train}/files/checkpoint"
    create_dir(f'{output_train}/patch_seg')

    train_dataset = PatchDataset(train_x, train_y)
    val_dataset = PatchDataset(val_x, val_y)

    train_loader = DataLoader(
        dataset=train_dataset,
        batch_size=config.batch_size,
        shuffle=True,
        num_workers=1)

    val_loader = DataLoader(
        dataset=val_dataset,
        batch_size=config.batch_size,
        shuffle=False,
        num_workers=1)

    device = torch.device(dev)
    model = VsegModel()

    model = model.to(device)

    optimizer = torch.optim.Adam(model.parameters(), lr=config.learning_rate)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', patience=5, verbose=True)
    loss_fn = DiceBCELoss()

    # Track model with Wandb
    wandb.watch(model, criterion=loss_fn, log="all", log_freq=25, log_graph=True)

    """ Training the model """
    train_losses = []
    val_losses = []

    best_valid_loss = float('inf')

    for epoch in (pbar := tqdm(range(config.epochs), desc="Epochs", leave=False, position=0)):
        train_loss, train_y, train_y_pred, x_train = train(model,
                                                           train_loader,
                                                           optimizer,
                                                           loss_fn,
                                                           device)

        val_loss, val_y, val_y_pred, x_val, f1_score, accuracy, jaccard, recall = evaluate(model,
                                                                                           val_loader,
                                                                                           loss_fn,
                                                                                           device)

        train_losses.append(train_loss)
        val_losses.append(val_loss)
        scheduler.step(val_loss)

        """ Saving model """
        if val_loss < best_valid_loss:
            best_epoch = epoch
            torch.save(model.state_dict(), f"{checkpoint_path}.latest.pth")
            best_valid_loss = val_loss

        if epoch % 10 == 0:
            torch.save(model.state_dict(), f"{checkpoint_path}.epoch_{epoch}.pth")

        pbar.set_postfix(
            loss=f'Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}')
        train_images = create_images(x_train, train_y, train_y_pred)
        val_images = create_images(x_val, val_y, val_y_pred)
        wandb.log({
            'Training Loss': train_loss,
            'Validation Loss': val_loss,
            'Accuracy': accuracy,
            'Jaccard': jaccard,
            'F1 score': f1_score,
            'Recall': recall,
            "Images": {"Training": [wandb.Image(x, mode="RGB") for x in train_images],
                       "Validation": [wandb.Image(x, mode="RGB") for x in val_images],
                       }
        })

        if epoch % 10 == 0:
            torch.save(model.state_dict(), f"{checkpoint_path}.epoch_{epoch}.pth")

    print(f'Finished Training: Best Epoch = {best_epoch}')
    artifact = wandb.Artifact('model', type='model')
    artifact.add_file(f"{checkpoint_path}.latest.pth")
    run.log_artifact(artifact)

    # For sweeps
    torch.save(model.state_dict(), os.path.join(wandb.run.dir, "model.pt"))
    run.finish()

    final_loss = pd.DataFrame(data=[train_losses, val_losses]).T
    final_loss.to_csv('final_metrics.csv', encoding='utf-8', index=False)


if __name__ == "__main__":
    # Hardcoded for wandb sweeps
    # dataset_dir = "/Users/frans/docker_data/confocal/vseg/patches"
    # output = "/Users/frans/docker_data/confocal/vseg/training"
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    dataset_dir = "/home/frans/code/vseg/data/patches"
    output = "/home/frans/code/vseg/data/training"

    train_model(dataset_dir=dataset_dir,
                dev=device,
                output_train=output)



import numpy as np
from skimage.io import imsave

from liom_toolkit.utils import load_zarr, load_node_by_name, generate_label_color_dict_allen
from .utils import convert_to_png_for_saving


def extract_single_slice_from_zarr(zarr_file: str, z: int, channel: int = 0, resolution_level: int = 0) -> np.ndarray:
    """
    Extracts a single slice from a 3D volume

    :param zarr_file: Path to the zarr file
    :type zarr_file: str
    :param z: Z index of the slice
    :type z: int
    :param channel: Channel to extract
    :type channel: int
    :param resolution_level: Resolution level to extract
    :type resolution_level: int
    :return: The extracted slice
    :rtype: np.ndarray
    """
    nodes = load_zarr(zarr_file)
    image_node = nodes[0]
    volume = image_node.data[resolution_level]
    if volume.ndim == 4:
        volume = volume[channel]

    image = volume[z, :, :]
    image = image.compute()
    return image


def extract_and_save_slice_form_zarr(zarr_file: str, z: int, data_dir: str, channel: int = 0,
                                     resolution_level: int = 0, name: str = "S1"):
    """
    Extracts a single slice from a 3D volume and saves it to disk

    :param zarr_file: Path to the zarr file
    :type zarr_file: str
    :param z: Z index of the slice
    :type z: int
    :param data_dir: Directory to save the slice
    :type data_dir: str
    :param channel: Channel to extract
    :type channel: int
    :param resolution_level: Resolution level to extract
    :type resolution_level: int
    :param name: Name of the volume
    :type name: str
    :return: The extracted slice
    :rtype: np.ndarray
    """
    image = extract_single_slice_from_zarr(zarr_file, z, channel, resolution_level)
    image = convert_to_png_for_saving(image)
    imsave(f"{data_dir}/{name}_C={channel}_Z={z}.png", image)

    return image


def extract_slices_form_zarr(zarr_file: str, start_z: int, num_slices: int, channel=0,
                             resolution_level=0) -> np.ndarray:
    """
    Extracts slices from a 3D volume

    :param zarr_file: Path to the zarr file
    :type zarr_file: str
    :param start_z: Z index of the slice
    :type start_z: int
    :param num_slices: Number of slices to extract
    :type num_slices: int
    :param channel: Channel to extract
    :type channel: int
    :param resolution_level: Resolution level to extract
    :type resolution_level: int
    :return: 3D volume with the extracted slices
    :rtype: np.ndarray
    """
    nodes = load_zarr(zarr_file)
    image_node = nodes[0]
    volume = image_node.data[resolution_level]
    if volume.ndim == 4:
        volume = volume[channel]

    image_zs = np.linspace(start_z - num_slices / 2, start_z + num_slices / 2, num_slices + 1, dtype=int)
    full_volume = np.zeros((len(image_zs), volume.shape[1], volume.shape[2]), dtype=np.uint32)

    for i, z in enumerate(image_zs):
        idx = int(z)
        image = volume[idx, :, :]
        image = image.compute()
        full_volume[i:, :, :] = image

    return full_volume


def extract_and_save_slices_form_zarr(zarr_file: str, start_z: int, num_slices: int, data_dir: str, channel: int = 0,
                                      resolution_level: int = 0, name: str = "S1",
                                      save_mip: bool = False) -> np.ndarray:
    """
    Extracts slices from a 3D volume and saves them to disk

    :param zarr_file: Path to the zarr file
    :type zarr_file: str
    :param start_z: Z index of the slice
    :type start_z: int
    :param num_slices: Number of slices to extract
    :type num_slices: int
    :param data_dir: Directory to save the slices
    :type data_dir: str
    :param channel: Channel to extract
    :type channel: int
    :param resolution_level: Resolution level to extract
    :type resolution_level: int
    :param name: Name of the volume
    :type name: str
    :param save_mip: Whether to save the maximum intensity projection
    :type save_mip: bool
    :return: 3D volume with the extracted slices
    :rtype: np.ndarray
    """
    volume = extract_slices_form_zarr(zarr_file, start_z, num_slices, channel=channel,
                                      resolution_level=resolution_level)

    imsave(f"{data_dir}/{name}_C={channel}_Z={start_z - num_slices // 2}-{start_z + num_slices // 2}.tif", volume)
    if save_mip:
        mip = np.max(volume, axis=0)
        mip = convert_to_png_for_saving(mip)
        imsave(f"{data_dir}/{name}_C={channel}_Z={start_z - num_slices // 2}-{start_z + num_slices // 2}_mip.png", mip)

    return volume


def colour_image(slice: np.ndarray, colour_dict: list):
    """
    Colour an image based on a colour dictionary

    :param slice: The slice to colour
    :type slice: np.ndarray
    :param colour_dict: The colour dictionary
    :type colour_dict: list
    :return: The coloured image
    :rtype: np.ndarray
    """
    slice_png = np.zeros_like(slice, dtype='uint8')

    # Add 3th dimension of size 3 to png
    slice_png = np.repeat(slice_png[:, :, np.newaxis], 3, axis=2)

    # Apply colour dict to image
    for i in range(len(colour_dict)):
        x, y = np.where(slice == colour_dict[i]['label-value'])
        slice_png[x, y, :] = colour_dict[i]['rgba'][0:3]

    return slice_png


def extract_label_slice(zarr_file: str, z: int, label: str, resolution_level: int = 0) -> np.ndarray:
    """
    Extracts a single slice from a 3D volume

    :param zarr_file: Path to the zarr file
    :type zarr_file: str
    :param z: Z index of the slice
    :type z: int
    :param label: The type of label
    :type label: str
    :param resolution_level: Resolution level to extract
    :type resolution_level: int
    :return: The extracted slice
    :rtype: np.ndarray
    """
    nodes = load_zarr(zarr_file)
    atlas_node = load_node_by_name(nodes, label)

    atlas = atlas_node.data[resolution_level]

    image = atlas[z, :, :]
    image = image.compute()
    return image


def save_atlas_slice(zarr_file: str, z: int, data_dir: str, resolution_level: int = 0, name: str = "S1") -> np.ndarray:
    """
    Extracts a single slice from a 3D volume and saves it to disk

    :param zarr_file: Path to the zarr file
    :type zarr_file: str
    :param z: Z index of the slice
    :type z: int
    :param data_dir: Directory to save the slice
    :type data_dir: str
    :param resolution_level: Resolution level to extract
    :type resolution_level: int
    :param name: Name of the volume
    :type name: str
    :return: The extracted slice
    :rtype: np.ndarray
    """
    image = extract_label_slice(zarr_file, z, "atlas", resolution_level)
    colour_dict = generate_label_color_dict_allen()
    image = colour_image(image, colour_dict)
    imsave(f"{data_dir}/{name}_Z={z}_atlas.png", image)

    return image


def save_vessel_slice(zarr_file: str, z: int, data_dir: str, colour_vessels: bool, resolution_level: int = 0,
                      name: str = "S1") -> np.ndarray:
    """
    Extracts a single vessel slice from a 3D volume and saves it to disk

    :param zarr_file: Path to the zarr file
    :type zarr_file: str
    :param z: Z index of the slice
    :type z: int
    :param data_dir: Directory to save the slice
    :type data_dir: str
    :param colour_vessels: Whether to colour the vessels according to the allen atlas
    :type colour_vessels: bool
    :param resolution_level: Resolution level to extract
    :type resolution_level: int
    :param name: Name of the volume
    :type name: str
    :return: The extracted slice
    :rtype: np.ndarray
    """
    image = extract_label_slice(zarr_file, z, "vessels", resolution_level)
    f_name = f"{data_dir}/{name}_Z={z}_vessels.png"
    if colour_vessels:
        colour_dict = generate_label_color_dict_allen()

        atlas_image = extract_label_slice(zarr_file, z, "atlas", resolution_level)
        coloured_vessels = image * atlas_image

        image = colour_image(coloured_vessels, colour_dict)
        f_name = f"{data_dir}/{name}_Z={z}_vessels_coloured.png"
    else:
        image = convert_to_png_for_saving(image)

    imsave(f_name, image)

    return image



import os
import random
from glob import glob

import matplotlib.pyplot as plt
import natsort
import numpy as np
import pandas as pd
import torch
from PIL import Image
from albumentations import HorizontalFlip, VerticalFlip, Rotate
from patchify import patchify
from skimage.color import gray2rgb
from skimage.exposure import equalize_adapthist
from skimage.io import imread, imsave
from skimage.transform import resize
from sklearn.metrics import accuracy_score, f1_score, jaccard_score, precision_score, recall_score
from tqdm.auto import tqdm


def matplotlib_imshow(img, one_channel=False):
    if one_channel:
        img = img.mean(dim=0)
    img = img / 2 + 0.5  # unnormalize
    npimg = img.numpy()
    if one_channel:
        plt.imshow(npimg, cmap="Greys")
    else:
        plt.imshow(np.transpose(npimg, (1, 2, 0)))


## Create a seed for reproducibility 
def seeding(seed):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True


## Creates a directory if it does not exist yet
def create_dir(path):
    if not os.path.exists(path):
        os.makedirs(path)


## Calculates time elapsed between start_time and end_time
def epoch_time(start_time, end_time):
    elapsed_time = end_time - start_time
    elapsed_mins = int(elapsed_time / 60)
    elapsed_secs = int(elapsed_time - (elapsed_mins * 60))
    return elapsed_mins, elapsed_secs


## Calculates relevant metrics between an instance of ground truth and prediction
## Metrics are F1, Recall, Precision, Accuracy, Jaccard
def calculate_metrics(y_true, y_pred):
    y_true = y_true > 0.5
    y_true = y_true.astype(np.uint8)
    y_true = y_true.reshape(-1)

    y_pred = y_pred > 0.5
    y_pred = y_pred.astype(np.uint8)
    y_pred = y_pred.reshape(-1)

    score_f1 = f1_score(y_true, y_pred)
    score_recall = recall_score(y_true, y_pred)
    score_precision = precision_score(y_true, y_pred)
    score_acc = accuracy_score(y_true, y_pred)
    score_jaccard = jaccard_score(y_true, y_pred)

    return [score_f1, score_recall, score_acc, score_jaccard, score_precision]


# Process an image to present to U-net model
def process_image(image, device):
    x = np.expand_dims(image, axis=0)
    x = np.expand_dims(x, axis=0)
    x = x.astype(np.float32)
    x = torch.from_numpy(x)
    x = x.to(device)
    return x


# Sort a list of filenames by numerical order
# This is used to order the patches as they are names 0_0_png, 1_0_png 2_0_png ... (instead of 0, 1, 10, 11 ...)
def numeric_filesort(path, folder="images", extension='png'):
    test = sorted(glob(f'{path}/{folder}/*{extension}'))
    test = natsort.natsorted(test, reverse=False)

    return test


# Add a inferred patch to empty array
def add_patch_to_empty_array(inference, pred_y, coords, stride, overlap, size, npatches):
    H = size[0]
    W = size[1]

    patch_x1 = coords[0] * stride
    patch_y1 = coords[1] * stride
    inference[patch_x1:(patch_x1 + H), patch_y1:(patch_y1 + W)] += pred_y

    if (coords[1] > 0 or coords[0] > 0) and overlap > 0:

        x1 = patch_x1
        y1 = patch_y1

        # If this is the first row (cannot be the first patch)
        if coords[0] == 0:
            x2 = x1 + H
            y2 = y1 + overlap
            to_add = [(x1, x2, y1, y2)]

        # If this is not the first row
        elif coords[0] > 0:
            # If this is the first column
            if coords[1] == 0:
                x2 = x1 + overlap
                y1 = y1
                y2 = y1 + W
                to_add = [(x1, x2, y1, y2)]

            # If this is between the first and last columns
            else:

                # This yields 2 rectangles:
                # rec1
                x1a = x1
                x2a = x1 + H
                y1a = y1
                y2a = y1 + overlap + 1

                # rec2
                x1b = x1
                x2b = x1 + overlap
                y1b = y1 + overlap + 1
                y2b = y1 + W
                to_add = [(x1a, x2a, y1a, y2a),
                          (x1b, x2b, y1b, y2b)]

        for rectangle in to_add:
            inference[rectangle[0]:rectangle[1], rectangle[2]:rectangle[3]] = inference[rectangle[0]:rectangle[1],
                                                                              rectangle[2]:rectangle[3]] / 2

    return (inference)


Image.MAX_IMAGE_PIXELS = None


def crop_image(image, size, stride):
    to_remove_x = (image.shape[1] - size[0]) % stride
    to_remove_left_x = np.floor(to_remove_x / 2).astype(int)
    to_remove_right_x = np.ceil(to_remove_x / 2).astype(int)

    to_remove_y = (image.shape[0] - size[0]) % stride
    to_remove_left_y = np.floor(to_remove_y / 2).astype(int)
    to_remove_right_y = np.ceil(to_remove_y / 2).astype(int)

    return (image[to_remove_left_y:image.shape[0] - to_remove_right_y,
            to_remove_left_x:image.shape[1] - to_remove_right_x])


def create_patches(image_path, size=(256, 256), stride=64, norm=False):
    patches = []

    image = imread(image_path, as_gray=True)
    image = crop_image(image, size, stride)
    image = (image / image.max() * 255).astype(np.uint8)

    if norm:
        image_clahe = equalize_adapthist(image, kernel_size=10, clip_limit=0.05, nbins=128)
    else:
        image_clahe = image

    patch = patchify(image_clahe, size, stride)

    patches.extend(
        patch.reshape(patch.shape[0] * patch.shape[1], patch.shape[2], patch.shape[3]))

    return patches, image.shape, patch.shape, image_clahe


def patch(image_path, save_path, norm, size=(256, 256), stride=64, augment=True, threshold=5, save_image=True,
          mask_path='0', use_mask=True, remove_background_tiles=False):
    image_name = image_path.split('/')
    image_name = image_name[len(image_name) - 1]
    image_name = image_name.replace('.png', '')

    img_train, shape, patch_shape, image = create_patches(image_path, size=size, stride=stride, norm=norm)

    if use_mask:
        mask_patch_path = image_path.replace('/image/', '/mask/')
        mask_patch_path = mask_patch_path.replace('.png', '_mask.png')
        mask_train, _, _, mask = create_patches(mask_patch_path, size=size, stride=stride, norm=False)

    pixel_sums = []
    name_idx = 0

    for i in tqdm(range(0, len(img_train)), desc='Patching image'):
        x = img_train[i]
        if remove_background_tiles and (np.abs(np.round(x.max(), 5) - np.round(x.min(), 5)) <= 0.001):
            continue

        if use_mask:
            y = mask_train[i]
        pixel_sums.append(img_train[i].sum())
        mean_pxl = img_train[i].sum() / (size[0] * size[1])
        if mean_pxl >= threshold:
            if augment:
                aug = HorizontalFlip(p=1.0)
                augmented = aug(image=x, mask=y)
                x1 = augmented['image']
                y1 = augmented['mask']
                aug = VerticalFlip(p=1.0)
                augmented = aug(image=x, mask=y)
                x2 = augmented['image']
                y2 = augmented['mask']
                aug = Rotate(limit=45, p=1.0)
                augmented = aug(image=x, mask=y)
                x3 = augmented['image']
                y3 = augmented['mask']
                X = [x, x1, x2, x3]
                Y = [y, y1, y2, y3]
            else:
                X = [x]
                if use_mask:
                    Y = [y]
            index = 0
            for i in X:
                i = resize(i, size)
                tmp_name = f"{image_name}_{name_idx}_{index}.png"
                create_dir(os.path.join(save_path, 'images'))
                image_path = os.path.join(save_path, 'images', tmp_name)
                if save_image:
                    i = gray2rgb(i, channel_axis=-1)
                    im = (i * 255).astype(np.uint8)
                    imsave(image_path, im, check_contrast=False)
                index += 1
            if use_mask:
                for m in Y:
                    tmp_name = f"{image_name}_{name_idx}_{index}.png"
                    create_dir(os.path.join(save_path, 'masks'))
                    mask_path = os.path.join(save_path, 'masks', tmp_name)
                    if save_image:
                        m = gray2rgb(m)
                        imsave(mask_path, m, check_contrast=False)
                    index += 1
            name_idx += 1
    output = pd.DataFrame(data=pixel_sums).T
    output.to_csv(f'{save_path}/pixel_sums.csv')

    if use_mask:
        return shape, patch_shape, image, mask
    else:
        return shape, patch_shape, image



import shutil

from .utils import *


def predict_one(model, img_path, save_path, stride=256, width=256, norm=True, dev="cuda", patching=True):
    """ Hyperparameters """
    if patching:
        H = width
        W = width
        size = (H, W)
        stride = stride
    else:
        image = imread(img_path)
        H = image.shape[0]
        W = image.shape[1]
        size = (H, W)
        stride = image.shape[0]

    device = torch.device(dev)

    image_name = img_path.split('/')
    image_name = image_name[len(image_name) - 1]
    image_id = image_name.replace('.png', '')

    overlap = W - stride

    create_dir(f'{save_path}')
    create_dir(f'{save_path}/patches')
    # Remove images if exists
    patches_images_dir = f'{save_path}/patches/images/'
    if os.path.exists(patches_images_dir):
        shutil.rmtree(patches_images_dir)
    create_dir(f'{save_path}/patches/images/')

    if patching:
        shape, patch_shape, processed_image = patch(img_path,
                                                    f'{save_path}/patches',
                                                    norm,
                                                    size,
                                                    stride,
                                                    augment=False,
                                                    threshold=0,
                                                    use_mask=False)
    else:
        # Only the clahe is done to the image
        image = imread(img_path)
        image = (image / image.max() * 255).astype(np.uint8)
        processed_image = equalize_adapthist(image, kernel_size=10, clip_limit=0.05, nbins=128)

        saved_image = gray2rgb(processed_image)
        saved_image = (saved_image / saved_image.max() * 255).astype(np.uint8)
        img_name = f"{image_id}_0_0.png"
        image_path = os.path.join(save_path, 'patches', 'images', img_name)
        imsave(image_path, saved_image, check_contrast=False)

    """ Load dataset """
    test_x = numeric_filesort(f'{save_path}/patches', folder="images")

    n_patches_by_row = (processed_image.shape[1] - W) / stride + 1

    x1 = 0
    y1 = 0
    inference = np.zeros(processed_image.shape)

    for x in test_x:
        image = imread(x, as_gray=True)
        image = process_image(image, device)
        image = image.to(device)
        with torch.no_grad():
            pred_y = model(image)
            pred_y = pred_y.cpu()
            pred_y = pred_y[0].numpy()
            pred_y = np.squeeze(pred_y, axis=0)
            pred_y = pred_y > 0.5
            pred_y = np.array(pred_y, dtype=np.uint8)
        if y1 % (n_patches_by_row) == 0 and y1 > 0:
            x1 += 1
            y1 = 0

        inference = add_patch_to_empty_array(inference,
                                             pred_y,
                                             [x1, y1],
                                             stride,
                                             overlap,
                                             size,
                                             n_patches_by_row)

        y1 += 1

    inference = np.floor(inference)
    inference = (inference / inference.max() * 255).astype(np.uint8)
    inference = inference.astype(bool)
    # inference = remove_small_objects(inference)
    inference = inference.astype(np.uint8) * 255

    save_inf = f"{save_path}/{image_id}_segmented.png"
    imsave(save_inf,
           inference, check_contrast=False)

    return inference



import cv2
import numpy as np
import torch
from torch.utils.data import Dataset


class PatchDataset(Dataset):
    def __init__(self, images_path, masks_path):
        self.images_path = images_path
        self.masks_path = masks_path
        self.n_samples = len(images_path)

    def __getitem__(self, index):
        image = cv2.imread(self.images_path[index], 0)
        image = image / 255.0
        image = np.expand_dims(image, axis=0)
        image = image.astype(np.float32)
        image = torch.from_numpy(image)

        mask = cv2.imread(self.masks_path[index], 0)
        mask = mask / 255.0
        mask = np.expand_dims(mask, axis=0)
        mask = mask.astype(np.float32)
        mask = torch.from_numpy(mask)

        return image, mask

    def __len__(self):
        return self.n_samples



import csv

import matplotlib.pyplot as plt
import numpy as np
from skimage.io import imread

from .cldice import cl_dice
from .predict_one import predict_one
from .utils import calculate_metrics


def show_diff(mask, prediction, output_path, id, acq) -> None:
    """
       - Black: TN
       - Red: FP
       - Blue: FN
       - White: TP
    """

    mask = mask > 0.5
    prediction = prediction > 0.5

    red = prediction * 1.0
    blue = mask * 1.0
    green = (prediction & mask) * 1.0

    rgb = np.stack([red, green, blue], axis=2)
    plt.imsave(f"{output_path}/{acq}_{id}_comparison.png", rgb)


def validate_model(model, img_list, save_path, device):
    # img_list: list of image paths to validate. The mask has to be in the same folder
    f1 = []
    recall = []
    accuracy = []
    jaccard = []
    cldice = []
    ids = []

    for images in img_list:
        image_name = images.split('/')
        image_id = image_name[len(image_name) - 1]
        image_id = image_id.replace('.png', '')
        ids.append(image_id)
        acquisition = image_name[len(image_name) - 2]

        inference = predict_one(model=model, img_path=images, save_path=save_path, norm=True, dev=device,
                                patching=False)

        mask_path = images.replace('.png', '_mask.png')
        mask = imread(mask_path)

        # comparison image
        mask = (mask / mask.max()).astype(np.uint8)
        inference = (inference / inference.max()).astype(np.uint8)
        show_diff(mask=mask, prediction=inference, output_path=save_path, id=image_id, acq=acquisition)

        # metrics
        [score_f1, score_recall, score_acc, score_jaccard, score_precision] = calculate_metrics(mask, inference)
        f1.append(score_f1)
        recall.append(score_recall)
        accuracy.append(score_acc)
        jaccard.append(score_jaccard)
        centerdice = cl_dice(inference, mask)
        cldice.append(centerdice)

    # averages
    f1_mean = sum(f1) / len(f1)
    recall_mean = sum(recall) / len(recall)
    accuracy_mean = sum(accuracy) / len(accuracy)
    jaccard_mean = sum(jaccard) / len(jaccard)
    cldice_mean = sum(cldice) / len(cldice)

    headings = ["Metrics"] + ids + ["mean"]
    accuracy_list = ["accuracy"] + accuracy + [accuracy_mean]
    f1_list = ["f1"] + f1 + [f1_mean]
    recall_list = ["recall"] + recall + [recall_mean]
    jaccard_list = ["jaccard"] + jaccard + [jaccard_mean]
    cldice_list = ["clDice"] + cldice + [cldice_mean]

    with open(f"{save_path}/validationmetrics.csv", mode="w") as f:
        csvwriter = csv.writer(f)
        csvwriter.writerow(headings)
        csvwriter.writerow(accuracy_list)
        csvwriter.writerow(f1_list)
        csvwriter.writerow(recall_list)
        csvwriter.writerow(jaccard_list)
        csvwriter.writerow(cldice_list)
